require 'pstore'

module Au
  class Repository
    # singleton class
    attr_reader :root, :path, :current, :head

    def self.root(path = nil)
      instance(path, false).root
    end

    def self.path(path = nil)
      instance(path, false).path
    end

    def self.head(path = nil)
      Commit.find instance(path, false).head
    end

    def self.diff(file_path)
      return nil unless head
      tracked_version = head.cat(file_path)
      return nil unless tracked_version
      `diff -c #{tracked_version.path} #{File.join(root, file_path)}`
    end

    def self.instance(path = nil, create = false)
      return @single_instance unless @single_instance.nil?
      begin
        @single_instance = new(path, create)
      rescue
        Kernel.abort("Fetal error: Repository not found. ")
      end
    end

    def initialize(path, create = false)
      if path.nil?
        working_dir = Dir.pwd
        # little bit different with mercurial 0.1
        # here we only check whether the current working directory is a repository
        if not File.directory?(File.join(working_dir, '.au')) and not create
          raise 'Repository not found in ' + working_dir
        end
        path = working_dir
      end

      @root = path
      @path = File.join(path, '.au')

      Dir.mkdir(@path) if create && !File.directory?(@path)

      @head_pstore = PStore.new(File.join(@path, 'head.pstore'))

      update_head_var

      @alt_pstore = PStore.new(File.join(@path, 'alt.pstore'))

      update_alt_list

      @leaves_pstore = PStore.new(File.join(@path, 'leaves.pstore'))
    end

    def merge(commit_to_merge)
      commit_id = Commit.merge(@head, commit_to_merge)
      return nil if commit_id.nil?
      # checkout to new commit
      checkout(commit_id)
      commit_id
    end

    def commit(commit_message)
      staged_file_paths = Staging.staged_file_paths
      return if staged_file_paths.empty?
      commit_created_id =
          if File.exist?(Repository.resolve_conflict_fname)
            parent_id1, parent_id2 = File.read(Repository.resolve_conflict_fname).split(',')
            File.delete(Repository.resolve_conflict_fname)
            Commit.create_commit_resolve_conflict(staged_file_paths, parent_id1, parent_id2)
          else
            Commit.create(staged_file_paths, commit_message, @head)
          end
      update_head_pstore(commit_created_id)
      update_leaves_pstore(commit_created_id)
      Staging.clear_staged_file_paths
      commit_created_id
    end

    def checkout(commit_id)
      this_commit = Commit.find(commit_id)
      if this_commit
        this_commit.checkout(update_head_var)
        update_alt_pstore(update_head_var)
        update_head_pstore(commit_id)
        update_head_var
      end
    end

    def pull(remote_repo)
      # Initialize local and remote db and commit list
      remote_commit_db = Commit.db(remote_repo.path)
      remote_commit_list = remote_commit_db.transaction do
        remote_commit_db.roots
      end
      local_commit_db = Commit.db
      local_commit_list = local_commit_db.transaction do
        local_commit_db.roots
      end
      # Check if the root commits are the same
      if local_commit_list[0] != remote_commit_list[0]
        raise "Error: Root commits do not match."
      end

      # Find commits that only exist on remote repo
      commit_id_to_add_list = remote_commit_list.reject{|x| local_commit_list.include?(x)}
      commits_to_add = {}
      remote_commit_db.transaction(true) do
        commit_id_to_add_list.each{|commit_id| commits_to_add[commit_id] = remote_commit_db[commit_id]}
      end

      # Write those commits to local commit db
      local_commit_db.transaction do
        commits_to_add.each{|commit_id, data| local_commit_db[commit_id] = data}
      end

      # Find documents and their diff_id in those commits that only existed remotely
      remote_diff_id_to_add_list = {}
      commits_to_add.each do |_, commit|
        commit[:doc_diff_ids].each do |doc_path, diff_id|
          remote_diff_id_to_add_list[doc_path] = diff_id
        end
      end

      # Write these information to local
      remote_diff_id_to_add_list.each do |doc_path, diff_id|
        diff_pstore_remote = PStore.new(diff_db_name_remote(doc_path, remote_repo.path))
        diff_pstore_remote.transaction(true) do
          diff_pstore_local = PStore.new(diff_db_name_remote(doc_path, @path))
          diff_pstore_local.transaction do
            if diff_pstore_local[diff_id] == nil
              diff_pstore_local[diff_id] = diff_pstore_remote[diff_id]
            elsif diff_pstore_local[diff_id] != diff_pstore_remote[diff_id]
              raise "Error: Remote and local reposiroty have different content in a same diff"
            end
          end
        end
      end

      # Find heads that only exist on remote repo
      heads_to_add = remote_repo.heads.reject{|x| heads.include?(x)}
      heads_to_add.each{|x| update_leaves_pstore(x)}
    end

    def push(remote_repo)
      # Initialize local and remote db and commit list
      remote_commit_db = Commit.db(remote_repo.path)
      remote_commit_list = remote_commit_db.transaction do
        remote_commit_db.roots
      end
      local_commit_db = Commit.db
      local_commit_list = local_commit_db.transaction do
        local_commit_db.roots
      end
      # Check if the root commits are the same
      if local_commit_list[0] != remote_commit_list[0]
        raise "Error: Root commits do not match."
      end

      # Find commits that only exist locally
      commit_id_to_add_list = local_commit_list.reject{|x| remote_commit_list.include?(x)}
      commits_to_add = {}
      local_commit_db.transaction(true) do
        commit_id_to_add_list.each{|commit_id| commits_to_add[commit_id] = local_commit_db[commit_id]}
      end

      # Write those commits to remote commit db
      remote_commit_db.transaction do
        commits_to_add.each{|commit_id, data| remote_commit_db[commit_id] = data}
      end

      # Find documents and their diff_id in those commits that only existed locally
      local_diff_id_to_add_list = {}
      commits_to_add.each do |_, commit|
        commit[:doc_diff_ids].each do |doc_path, diff_id|
          local_diff_id_to_add_list[doc_path] = diff_id
        end
      end

      # Write these information to remote
      local_diff_id_to_add_list.each do |doc_path, diff_id|
        diff_pstore_local = PStore.new(diff_db_name_remote(doc_path, @path))
        diff_pstore_local.transaction(true) do
          diff_pstore_remote = PStore.new(diff_db_name_remote(doc_path, remote_repo.path))
          diff_pstore_remote.transaction do
            if diff_pstore_remote[diff_id] == nil
              diff_pstore_remote[diff_id] = diff_pstore_local[diff_id]
            elsif diff_pstore_local[diff_id] != diff_pstore_remote[diff_id]
              raise "Error: Remote and local reposiroty have different content in a same diff"
            end
          end
        end
      end
      # Find heads that only exist on remote repo
      heads_to_add = heads.reject{|x| remote_repo.heads.include?(x)}
      heads_to_add.each{|x| remote_repo.update_leaves_pstore(x)}
    end

    # def clone
    #
    # end

    def heads
      @leaves_pstore.transaction(true) do
        @leaves_pstore.roots
      end
    end

    def update_head_var
      @head = @head_pstore.transaction(true) do
        @head_pstore[:head]
      end
    end

    def update_leaves_pstore(commit_id)
      @leaves_pstore.transaction do
        commit = Commit.find(commit_id)
        if commit
          parent_commit_id_1, parent_commit_id_2 = commit.parent_id_1, commit.parent_id_2
          @leaves_pstore.delete(parent_commit_id_1)
          @leaves_pstore.delete(parent_commit_id_2)
          @leaves_pstore[commit_id] = 1
        end
      end
    end

    private
    def diff_db_name_remote(file_path, remote_path)
      paths = [remote_path, "diff_#{file_path.gsub('/', '.')}.pstore"].compact
      File.join(*paths)
    end

    def update_head_pstore(commit_id)
      @head = @head_pstore.transaction do
        @head_pstore[:head] = commit_id
      end
    end

    def update_alt_list
      @alt_list = @alt_pstore.transaction(true) do
        @alt_pstore[:commit]
      end
    end

    def update_alt_pstore(commit_id)
      @alt_pstore.transaction do
        @alt_pstore[:commit].nil? ? (@alt_pstore[:commit] = [commit_id]) : (@alt_pstore[:commit] << commit_id)
      end
    end

    def self.resolve_conflict_fname
      File.join(Repository.instance.path, 'resolve_conflict')
    end
  end
end
